#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/syslog.h>

#include "config.h"
#include "libhttpd.h"

static httpd_server* hs = (httpd_server*) 0;
int terminate = 0;
static connecttab* connects;
int numconnects = 0, cnum;

static int maxconnects = MAXCLIENT;
char* user = SERVER_USER;
char* dir = DOCUMENT_ROOT;


static void handle_term( int sig )
{
	if ( hs != (httpd_server*) 0 )
		free_httpd_server( hs );

	for ( cnum = 0; cnum < maxconnects; cnum++ )
	{
		if ( connects[cnum].hc->file > 0 )
			close( connects[cnum].hc->file );
		free( connects[cnum].hc );
	}
	free( connects );
	exit( 0 );
}

int main( int argc, char** argv )
{
	httpd_sockaddr sa;
	fd_set active_fd_set, read_fd_set;
	int fd_max;
	int client_status;
	struct timeval tv;
	time_t now;

	int tv_sec = SELECT_TV_SEC;
	int tv_usec = SELECT_TV_USEC;

	sa.sa.sa_family = AF_INET;
	sa.sa_in.sin_addr.s_addr = htonl( INADDR_ANY );
	sa.sa_in.sin_port = htons( PORT );

	(void) signal( SIGTERM, handle_term );
	(void) signal( SIGINT, handle_term );
	(void) signal( SIGPIPE, SIG_IGN );		  // get EPIPE instead
	(void) signal( SIGHUP, handle_term );
	(void) signal( SIGUSR1, handle_term );

	hs = httpd_initialize( &sa, PORT );
	if ( hs == (httpd_server*) 0 )
		exit( 1 );

	connects = NEW( connecttab, maxconnects );
	if ( connects == (connecttab*) 0 )
	{
		syslog( LOG_CRIT, "out of memory" );
		exit( 1 );
	}

	for ( cnum = 0; cnum < maxconnects; cnum++ )
	{
		connects[cnum].conn_state = CNST_FREE;
		connects[cnum].hc = NEW( httpd_conn, 1 );
		conn_first_init( connects[cnum].hc );
	}

	FD_ZERO(&active_fd_set);
	FD_SET(hs->listen_fd, &active_fd_set);
	fd_max = hs->listen_fd+1;

	while ( ( ! terminate ) || numconnects > 0 )
	{
		time( &now );
		tv.tv_sec = tv_sec;
		tv.tv_usec = tv_usec;
		read_fd_set = active_fd_set;
		if (select ( fd_max, &read_fd_set, NULL, NULL, &tv ) < 0)
		{
			syslog( LOG_CRIT, "select error" );
			continue;
		}

		if ( numconnects > 0 )
		{
			fd_max = hs->listen_fd+1;
			for ( cnum = 0; cnum < maxconnects; cnum++ )
			{
				switch ( connects[cnum].conn_state )
				{
					case CNST_READING :
						if ( FD_ISSET( connects[cnum].hc->conn_fd, &read_fd_set ) )
						{
							client_status = httpd_got_request( connects[cnum].hc );
							if ( client_status < 0 )
							{
								FD_CLR(connects[cnum].hc->conn_fd, &active_fd_set);
								clear_client( connects, cnum, &numconnects );
							}else if ( client_status == 1 ) {
								FD_CLR(connects[cnum].hc->conn_fd, &active_fd_set);
								connects[cnum].conn_state = CNST_SENDING;
							}else{
								if ( ( now - TIMEOUT ) >= connects[cnum].hc->now )
								{
									FD_CLR(connects[cnum].hc->conn_fd, &active_fd_set);
									clear_client( connects, cnum, &numconnects );
								}else if ( fd_max <= connects[cnum].hc->conn_fd ) fd_max = connects[cnum].hc->conn_fd+1;
							}
						}else{
							if ( ( now - TIMEOUT ) >= connects[cnum].hc->now )
							{
								FD_CLR(connects[cnum].hc->conn_fd, &active_fd_set);
								clear_client( connects, cnum, &numconnects );
							}else if ( fd_max <= connects[cnum].hc->conn_fd ) fd_max = connects[cnum].hc->conn_fd+1;
						}
						break;
					case CNST_SENDING :
						client_status = httpd_start_request( connects[cnum].hc );
						if ( client_status == -1 )
						{
							FD_CLR(connects[cnum].hc->conn_fd, &active_fd_set);
							clear_client( connects, cnum, &numconnects );
						}else if ( client_status == 1 ) {
							if ( connects[cnum].hc->keep_alive == 0 )
							{
								FD_CLR(connects[cnum].hc->conn_fd, &active_fd_set);
								clear_client( connects, cnum, &numconnects );
							}else{
								conn_init( connects[cnum].hc );
								connects[cnum].conn_state = CNST_READING;
								if ( fd_max <= connects[cnum].hc->conn_fd ) fd_max = connects[cnum].hc->conn_fd+1;
							}
						}
						break;
				}
			}
		}
		if ( FD_ISSET (hs->listen_fd, &read_fd_set) )
		{
			for ( cnum = 0; cnum < maxconnects; cnum++ )
				if ( connects[cnum].conn_state == CNST_FREE )
					break;
			if ( maxconnects  == cnum )
				continue;
			numconnects++;
			connects[cnum].conn_state = CNST_READING;
			if ( httpd_get_conn( hs->listen_fd, connects, cnum ) < 0 )
			{
				clear_client( connects, cnum, &numconnects );
				continue;
			}
			FD_SET( connects[cnum].hc->conn_fd, &active_fd_set);
			if ( fd_max <= connects[cnum].hc->conn_fd ) fd_max = connects[cnum].hc->conn_fd+1;
			continue;
		}
	}

}
